# -*- coding: utf-8 -*-
import os

from cronq.utils import to_bool


class Config(object):
    BUGSNAG_API_KEY = os.getenv('BUGSNAG_API_KEY')
    CRONQ_QUEUE = os.getenv('CRONQ_QUEUE', 'cronq_jobs')
    DATABASE_URL = os.getenv(
        'DATABASE_URL',
        os.getenv('CRONQ_MYSQL',
                  'mysql+pymysql://root@localhost/cronq')).replace(
                        'mysql://', 'mysql+pymysql://')
    DEBUG = to_bool(os.getenv('DEBUG', 0))
    LISTEN_INTERFACE = os.getenv('LISTEN_INTERFACE', '0.0.0.0')
    LOG_PATH = os.getenv('LOG_PATH', '/var/log/cronq') if not os.path.exists('/var/log/cronq') else '/tmp'  # noqa
    PORT = int(os.getenv('PORT', 5000))
    RABBITMQ_URL = os.getenv('RABBITMQ_URL', 'amqp://guest:guest@localhost/')
    SECRET_KEY = os.getenv('SECRET_KEY', 'development key')
    SENTRY_DSN = os.getenv('SENTRY_DSN')



# -*- coding: utf-8 -*-
__version__ = '0.26.2'

# import here
import cronq.logsetup


def make_application():
    from flask import Flask
    from cronq.config import Config
    import cronq.web
    import os

    flask_app = Flask(__name__,
                      static_url_path='/static')

    if Config.BUGSNAG_API_KEY:
        import bugsnag
        from bugsnag.flask import handle_exceptions
        bugsnag.configure(api_key=Config.BUGSNAG_API_KEY)
        handle_exceptions(flask_app)
    elif Config.SENTRY_DSN:
        from raven.contrib.flask import Sentry
        sentry = Sentry()
        sentry.init_app(flask_app, dsn=Config.SENTRY_DSN)

    flask_app.config.from_object('cronq.config.Config')
    flask_app.register_blueprint(cronq.web.blueprint_http)
    cronq.web.blueprint_http.config = flask_app.config

    return flask_app



# -*- coding: utf-8 -*-
import datetime
import json
import logging

from cronq import interval_parser
from cronq.backends.mysql import Storage
from cronq.utils import json_serial
from cronq.utils import query_category_id
from cronq.utils import query_category_name
from cronq.utils import query_id
from cronq.utils import query_page
from cronq.utils import query_per_page
from cronq.utils import query_sort
from cronq.utils import split_command
from cronq.utils import task_status
from cronq.utils import took

from flask import Blueprint
from flask import Response
from flask import abort
from flask import flash
from flask import g
from flask import redirect
from flask import render_template
from flask import request
from flask import url_for

blueprint_http = Blueprint('blueprint_http', __name__)
blueprint_http.add_app_template_filter(split_command, 'split_command')
blueprint_http.add_app_template_filter(task_status, 'task_status')
blueprint_http.add_app_template_filter(took, 'took')

logger = logging.getLogger(__name__)


@blueprint_http.before_request
def create_storage():
    if request.path.startswith('/static/'):
        return
    g.storage = Storage(isolation_level=None)


@blueprint_http.after_request
def remove_storage(request):
    if hasattr(g, 'storage'):
        try:
            g.storage.close()
        except Exception:
            logger.exception("exception in remove storage")
    return request


@blueprint_http.route('/')
def index():
    jobs = list(g.storage.jobs())
    categories = list(g.storage.categories())
    categories = {category['id']: category for category in categories}
    return render_template('index.html', jobs=jobs, categories=categories)


@blueprint_http.route('/_status')
def status():
    return Response(
        json.dumps({'status': 'OK'}),
        mimetype='application/json',
    )


@blueprint_http.route('/job/<int:id>', methods=['GET', 'POST'])
def job(id):
    if request.method == 'POST' and request.form.get('run_now') is not None:
        g.storage.run_job_now(id)
        flash('Job submitted at {0}'.format(datetime.datetime.utcnow()))
        return redirect(url_for('.job', id=id))

    job_doc = g.storage.get_job(id)
    chunks = g.storage.last_event_chunks_for_job(id, 20)
    title = job_doc.get('name', '')
    return render_template('job.html', job=job_doc, chunks=chunks, title=title)


@blueprint_http.route('/run/<string:id>')
def run_id(id):
    events = list(g.storage.events_for_run_id(id))
    job_id = events[0]['job_id']
    job = g.storage.get_job(job_id)
    return render_template('run_id.html', events=events, job=job)


@blueprint_http.route('/failures')
def failures():
    failure_events = list(g.storage.failures())
    names = {job['id']: job['name'] for job in g.storage.jobs()}
    for event in failure_events:
        event['job_name'] = names[event['job_id']]
    return render_template('failures.html', events=failure_events)


@blueprint_http.route('/api/category/<string:name>', methods=['PUT', 'POST'])
def category(name):
    data = request.json
    logger.info("Retrieving jobs")
    existing_jobs = g.storage.jobs_for_category(name=name)

    logger.info("Retrieving category")
    category_id = g.storage.category_id_for_name(name)
    job_lookup = {}

    logger.info("Validating jobs")
    if not validate_unique_job_names(data.get('jobs', [])):
        abort(400)

    logger.info("Indexing existing jobs")
    for job in existing_jobs:
        job_lookup[job['name']] = job

    logger.info("Processing posted jobs")
    for job in data.get('jobs', []):
        name = job['name']
        logger.info("Calcuating next run for {0}".format(name))
        next_run, duration = interval_parser.next_run_and_duration_from_8601(
            job['schedule'])
        existing_job = job_lookup.get(name, {})
        new_id = existing_job.get('id')
        new_interval = duration.total_seconds()
        command = job['command']

        logger.info("Adding job {0}".format(name))
        g.storage.add_job(
            name,
            new_interval,
            command,
            next_run,
            new_id,
            category_id,
            routing_key=job.get('routing_key')
        )
        if existing_job:
            del job_lookup[name]

    logger.info("Removing old jobs: {0}".format(job_lookup.keys()))
    remove_jobs(g.storage, job_lookup.itervalues())

    return '{"status": "success"}'


@blueprint_http.route('/api/categories', methods=['GET'])
def api_categories():
    _id = query_id(request.args)
    categories = g.storage.categories(_id=_id)
    return Response(
        json.dumps({
            'data': {
                'categories': list(categories)
            },
        }, default=json_serial),
        mimetype='application/json',
    )


@blueprint_http.route('/api/categories/<string:name>', methods=['GET'])
def api_category_show(name):
    category = g.storage.categories_first(name)
    return Response(
        json.dumps({
            'data': {
                'category': {
                    'id': category.id,
                    'name': category.name,
                },
            },
        }, default=json_serial),
        mimetype='application/json',
    )


@blueprint_http.route('/api/jobs', methods=['GET'])
def api_jobs():

    per_page = query_per_page(request.args)
    page = query_page(request.args)
    sort = query_sort(request.args, allowed_fields=['id', 'name', 'category_id'])
    category_id = query_category_id(request.args)
    _id = query_id(request.args)

    category_name = query_category_name(request.args)
    if category_name:
        category = g.storage.categories_first(category_name)
        if category is None:
            return Response(
                json.dumps({
                    'data': {
                        'jobs': [],
                    },
                }, default=json_serial),
                mimetype='application/json'
            )
        category_id = category.id

    jobs = g.storage.jobs(
        _id=_id,
        category_id=category_id,
        page=page,
        per_page=per_page,
        sort=sort,
        include_runs=True)

    return Response(
        json.dumps({
            'data': {
                'jobs': list(jobs),
            },
        }, default=json_serial),
        mimetype='application/json'
    )


@blueprint_http.route('/api/jobs/<int:id>', methods=['GET'])
def api_job_show(id):
    jobs = list(g.storage.jobs(_id=id, per_page=1, include_runs=True))
    if len(jobs) != 1:
        return Response(
            json.dumps({
                'error': {
                    'message': 'Job not found for id {0}'.format(id),
                },
            }, default=json_serial),
            mimetype='application/json'
        )

    return Response(
        json.dumps({
            'data': {
                'job': job[0],
            },
        }, default=json_serial),
        mimetype='application/json'
    )


@blueprint_http.route('/api/jobs/<int:id>/run', methods=['POST'])
def api_job_run(id):
    jobs = list(g.storage.jobs(_id=id, per_page=1))

    if len(jobs) != 1:
        return Response(
            json.dumps({
                'error': {
                    'message': 'Job not found for id {0}'.format(id),
                },
            }, default=json_serial),
            mimetype='application/json'
        )

    g.storage.run_job_now(id)
    return Response(
        json.dumps({
            'success': {
                'message': 'Job submitted at {0}'.format(datetime.datetime.utcnow()),
            },
        }, default=json_serial),
        mimetype='application/json'
    )


def remove_jobs(storage, jobs):
    for job in jobs:
        g.storage.remove_job(job['id'])


def validate_unique_job_names(jobs):
    job_names = [job['name'] for job in jobs]
    return len(job_names) == len(set(job_names))



# -*- coding: utf-8 -*-
import argparse
import logging

from cronq.backends.mysql import Storage
from cronq.models.event import Event
from sqlalchemy import between

logger = logging.getLogger(__name__)


def prune_record(event_id, max_event_id, storage):
    last = event_id + 100
    if last > max_event_id:
        last = max_event_id

    stmt = Event.__table__.delete().where(between(Event.id, event_id, last))
    storage._engine.execute(stmt)
    logger.info('Pruning {0} - {1}'.format(event_id, last))
    storage.session.commit()
    return last


def prune(first, last):
    storage = Storage(isolation_level=None)

    event_id = first
    while event_id <= last:
        try:
            event_id = prune_record(event_id, last, storage)
        except (KeyboardInterrupt, SystemExit):
            storage.session.commit()
            return
        except Exception as e:
            logger.warning(e)
            return

    storage.session.commit()


def main():
    parser = argparse.ArgumentParser(description='Prunes the cronq datastore')
    parser.add_argument('--first',
                        type=int,
                        default='0',
                        help='first entry to prune')
    parser.add_argument('--last',
                        type=int,
                        default='0',
                        help='last entry to prune')
    args = parser.parse_args()
    args = vars(args)
    prune(args['first'], args['last'])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import datetime


def split_command(string):
    commands = string.strip().split(';')
    ret = []
    for command in commands:
        ret.extend(command.strip().split(' && '))
    return ret


def task_status(first, last=None):
    if last is None:
        return first['status']

    if last['status'] == 'finished':
        if int(last['return_code']) == 0:
            return 'succeeded'
        return 'failed'
    return last['status']


def took(first_time, last_time):
    if type(first_time) is not datetime.datetime:
        return ""

    if type(last_time) is not datetime.datetime:
        return ""

    elapsed_time = last_time - first_time
    return int(elapsed_time.total_seconds())


def unicodedammit(x):
    """encode as string, decode as unicode bytes"""
    try:
        decoded = decode_string(x, ignore=False)
        return decoded
    except Exception:
        # decide with ignore
        decoded = decode_string(x, ignore=True)
        return decoded
    return decoded


def decode_string(newstr, ignore=False):
    """attempt to decode the string with three different options
        ascii
        utf8mb4
        utf8
        latin1
    If none of these works raise an error
    """

    # this is already unicode. return
    if type(newstr) is unicode:
        return newstr

    encodings = ['ascii', 'utf8mb4', 'utf8', 'latin1']
    for encoding in encodings:
        try:
            if ignore:
                dec_str = newstr.decode(encoding, 'ignore')
            else:
                dec_str = newstr.decode(encoding)
            return dec_str
        except:
            pass

    raise Exception("can't decode this string at all")


def chunks_to_runs(chunks):
    runs = []
    for chunk in chunks:
        run_id = None
        host = None
        for entry in ['first', 'last']:
            if chunk.get(entry, None) is None:
                continue

            if run_id is None:
                run_id = chunk.get(entry, {}).get('run_id', None)
            if host is None:
                host = chunk.get(entry, {}).get('host', None)

        status = 'pending'
        completed_at = None
        return_code = None
        completed_event_id = None
        if chunk.get('last', {}):
            status = chunk.get('last', {}).get('status', 'pending')
            completed_at = chunk.get('last', {}).get('datetime', None)
            return_code = chunk.get('last', {}).get('return_code', None)
            completed_event_id = chunk.get('last', {}).get('id', None)

        started_at = None
        started_event_id = None
        first = chunk.get('first', None)
        if first is not None:
            started_at = chunk.get('first', {}).get('datetime', None)
            started_event_id = chunk.get('first', {}).get('id', None)

        runs.append({
            'id': run_id,
            'job_id': chunk.get('job_id'),
            'status': status,
            'completed_at': completed_at,
            'completed_event_id': completed_event_id,
            'started_at': started_at,
            'started_event_id': started_event_id,
            'return_code': return_code,
            'host': host,
        })

    return runs


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if hasattr(obj, 'isoformat'):
        serial = obj.isoformat()
        return serial
    if isinstance(obj, datetime.timedelta):
        return obj.total_seconds()
    raise TypeError ("Type %s not serializable" % type(obj))


def query_id(args):
    _id = args.get('id', None)
    if _id is not None:
        _id = int(_id)

    return _id


def query_category_id(args):
    category_id = args.get('category.id')
    if category_id is not None:
        category_id = int(category_id)

    return category_id


def query_category_name(args):
    return args.get('category.name', None)


def query_page(args):
    page = int(args.get('page', 0))
    if page < 0:
        page = 0

    return page


def query_per_page(args):
    per_page = int(args.get('per_page', 10))
    if per_page > 100:
        per_page = 100

    return per_page


def query_sort(args, allowed_fields=None):
    field, order = args.get('sort', 'id.asc').split('.', 1)
    if not order:
        order = 'asc'

    if allowed_fields is None:
        allowed_fields = ['id', 'name']

    if field not in allowed_fields:
        field = 'id'

    return '{0}.{1}'.format(field, order)


def to_bool(s):
    try:
        int_s = int(s)
        return bool(int_s) is True
    except:
        pass

    if isinstance(s, basestring):
        if s.lower() in ['true', 't']:
            return True
        if s.lower() in ['false', 'f']:
            return False

    return bool(s) is True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import logging
import json
import jsonschema
import sys


def cronq_schema():
    return {
        "title": "cronq",
        "type": "object",
        "properties": {
            "category": {
                "type": "string"
            },
            "jobs": {
                "type": "array",
                "items": {
                    "title": "job",
                    "type": "object",
                    "properties": {
                        "name": {
                            "type": "string",
                        },
                        "schedule": {
                            "type": "string",
                            "pattern": "^R/[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2})?/(P([0-9]+Y)?([0-9]+M)?([0-9]+D)?T?([0-9]+H)?([0-9]+M)?([0-9]+S)?|P[0-9]+W)$",  # noqa
                        },
                        "command": {
                            "type": "string",
                        },
                        "routing_key": {
                            "type": "string",
                        },
                    },
                    "required": [
                        "name",
                        "schedule",
                        "command",
                        "routing_key",
                    ],
                },
            },
        },
        "required": [
            "category",
            "jobs",
        ],
    }


def validate(config_file):
    logger = logging.getLogger('amqp-dispatcher')
    logger.setLevel(logging.ERROR)

    config = None
    with open(config_file) as f:
        config = json.loads(f.read())

    if config is None:
        logger.error('Failed to open config file')
        return False

    jsonschema.validate(config, cronq_schema())
    return True


def main():
    parser = argparse.ArgumentParser(description='Validate cronq.config')
    parser.add_argument('--config',
                        metavar='config',
                        type=str,
                        default='cronq.config',
                        help='path to the config file')
    args = parser.parse_args()
    if not validate(args.config):
        sys.exit(1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import datetime
import logging
import os
import socket
import urllib

from cronq.config import Config
from cronq.models.category import Category
from cronq.models.event import Event
from cronq.models.job import Job
from cronq.utils import chunks_to_runs

from sqlalchemy import create_engine
from sqlalchemy import or_
from sqlalchemy.exc import InternalError
from sqlalchemy.exc import OperationalError
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.expression import asc
from sqlalchemy.sql.expression import desc

from uuid import uuid4


logger = logging.getLogger(__name__)


class Storage(object):

    FINISHED = 'finished'
    FAILED = 'failed'
    STARTED = 'started'
    SUCCEEDED = 'succeeded'

    def __init__(self, publisher=None, isolation_level='SERIALIZABLE'):
        self.publisher = publisher

        self._engine = self._new_engine(isolation_level=isolation_level)
        self._maker = sessionmaker(bind=self._engine)
        self.session = self._new_session()

    def _new_engine(self, isolation_level='SERIALIZABLE'):
        if isolation_level is not None:
            return create_engine(Config.DATABASE_URL,
                                 isolation_level=isolation_level)
        return create_engine(Config.DATABASE_URL)

    def _new_session(self):
        return self._maker()

    def bootstrap(self):
        models = [
            Category,
            Job,
            Event,
        ]
        for model in models:
            try:
                model.__table__.create(self._engine)
            except (InternalError, OperationalError, ProgrammingError):
                logger.debug("error in create table - probably already exists: {}".format(model.__table__))

    def close(self):
        self.session.close()
        self._engine.dispose()

    def add_job(self,
                name,
                interval_seconds,
                command,
                next_run,
                id=None,
                category_id=None,
                routing_key=None):
        if routing_key is None:
            routing_key = 'default'
        job = Job()
        job.id = id
        job.name = name
        job.next_run = next_run
        job.interval = datetime.timedelta(seconds=interval_seconds)
        job.command = command
        job.routing_key = routing_key
        job.category_id = category_id
        self.session.merge(job)
        self.session.commit()

    def remove_job(self, job_id):
        job = self.session.query(Job).get(job_id)
        if job:
            self.session.delete(job)
            self.session.commit()

    def add_event(self, job_id, _datetime, run_id, type, host, return_code):
        event = Event()
        event.job_id = job_id
        event.datetime = _datetime
        event.run_id = run_id
        event.type = type
        event.host = host
        event.return_code = return_code
        self.session.add(event)
        self.session.commit()

    def update_job_status(self, run_id, job_id, _datetime, status, return_code=None):
        while True:
            try:
                self._update_job_status(job_id, _datetime, status, return_code)
                break
            except InternalError:
                logger.warning('[cronq_job_id:{0}] [cronq_run_id:{1}] Unable to update job with result, retrying'.format(
                    job_id, run_id
                ))
                self.session.rollback()
        return True

    def _update_job_status(self, job_id, _datetime, new_status, return_code=None):
        job = self.session.query(Job).get(job_id)

        if not job:
            logger.warning('[cronq_job_id:{0}] job not found during update')
            return

        # update record if found
        job.current_status = new_status
        if new_status == self.STARTED:
            job.last_run_start = _datetime
            job.last_run_status = None
            job.last_run_stop = None
        if new_status == self.FAILED:
            job.last_run_status = new_status
            job.last_run_stop = _datetime
        if new_status == self.FINISHED:
            job.last_run_status = new_status
            job.last_run_stop = _datetime
            if return_code is not None:
                job.current_new_status = self.FAILED
                job.last_run_status = self.FAILED
                if int(return_code) == 0:
                    job.current_status = self.SUCCEEDED
                    job.last_run_status = self.SUCCEEDED
        self.session.merge(job)
        self.session.commit()

    def jobs(self, _id=None, category_id=None, page=0, per_page=None, sort='category_id.asc', include_runs=False):
        session = self.session
        categories = list(self.categories())

        query = session.query(Job)

        if _id is not None:
            query = query.filter_by(id=_id)

        if category_id is not None:
            query = query.filter_by(category_id=category_id)

        field, order = sort.split('.', 1)
        if order == 'asc':
            query = query.order_by(asc(getattr(Job, field)))
            if field != 'id':
                query = query.order_by(asc(Job.id))
        elif order == 'desc':
            query = query.order_by(desc(getattr(Job, field)))
            if field != 'id':
                query = query.order_by(asc(Job.id))

        if per_page is not None:
            query = query.limit(per_page)

        if page > 0:
            query = query.offset(page * per_page)

        for job in query:
            data = {
                'id': job.id,
                'category_id': job.category_id,
                'name': job.name,
                'interval': job.interval,
                'next_run': job.next_run,
                'last_run_start': job.last_run_start,
                'last_run_stop': job.last_run_stop,
                'last_run_status': job.last_run_status,
                'last_run_text': job.last_run_text(),
                'current_status': job.current_status or 'none',
                'routing_key': job.routing_key,
                'command': job.command,
                'locked_by': job.locked_by,
            }
            if include_runs:
                data['runs'] = chunks_to_runs(self.last_event_chunks_for_job(job.id, 20))
            category = filter(lambda c: c['id'] == job.category_id, categories)
            if len(category) == 1:
                data['category'] = category[0]
            yield data

    def categories(self, _id=None, sort='id.asc'):
        session = self.session

        query = session.query(Category)

        if _id is not None:
            query = query.filter_by(id=_id)

        field, order = sort.split('.', 1)
        if order == 'asc':
            query = query.order_by(asc(getattr(Category, field)))
        elif order == 'desc':
            query = query.order_by(desc(getattr(Category, field)))

        for category in query:
            yield {
                'id': category.id,
                'name': category.name,
            }

    def categories_first(self, name):
        session = self.session
        return session.query(Category).filter_by(name=name).first()

    def get_job(self, id):
        job = self.session.query(Job).get(id)
        return self._job_doc_for_model(job)

    def _job_doc_for_model(self, job):
        return {
            'id': job.id,
            'name': job.name,
            'next_run': job.next_run,
            'interval': job.interval,
            'command': job.command,
        }

    def jobs_for_category(self, id=None, name=None):
        if id and name:
            assert "Don't pass both id and name"

        if name:
            id = self.category_id_for_name(name)

        for job in self.session.query(Job).filter_by(category_id=id):
            yield self._job_doc_for_model(job)

    def category_id_for_name(self, name):
        category = self.session.query(Category).filter_by(name=name).first()
        if category is None:
            category = Category(name=name)
            self.session.add(category)
            self.session.commit()
        return category.id

    def last_event_chunks_for_job(self, job_id, number):
        events = self.session.query(Event).filter_by(job_id=job_id).\
            order_by(desc(Event.id)).limit(number)
        events = [doc for doc in self.event_models_to_docs(events)]

        if len(events) == 0:
            return []


        chunks = {}
        for event in events:
            if event['run_id'] not in chunks:

                chunks[event['run_id']] = {
                    'first': None,
                    'last': None,
                    'job_id': event['job_id']
                }

            if event['type'] == 'starting':
                chunks[event['run_id']]['first'] = event
            else:
                chunks[event['run_id']]['last'] = event

        log_url_template = os.getenv('CRONQ_LOG_URL_TEMPLATE', None)
        if log_url_template:
            self._add_log_urls(chunks, log_url_template)

        docs = [chunk for i, chunk in chunks.iteritems()]

        def chunk_compare(x, y):
            first = x.get('first', x.get('last'))
            last = y.get('first', y.get('last'))
            if first is None and last is None:
                return 0
            if first is None:
                return -1
            if last is None:
                return 1
            return int((first['datetime'] - last['datetime']).total_seconds())

        # newest to oldest
        docs = sorted(docs, cmp=chunk_compare, reverse=True)

        return docs

    def events_for_run_id(self, run_id):
        events = self.session.query(Event).filter_by(run_id=run_id).\
            order_by(Event.id)
        return self.event_models_to_docs(events)

    def failures(self):
        events = self.session.query(Event)\
            .filter_by(type=self.FINISHED)\
            .filter(Event.return_code != 0)\
            .order_by(desc(Event.id)).limit(50)
        return self.event_models_to_docs(events)

    def run_job_now(self, id):
        event = self.session.query(Job).get(id)
        event.run_now = True
        self.session.merge(event)
        self.session.commit()

    def inject(self):
        """Get a message from storage and injects it into the job stream"""
        logger.info('Trying to inject')
        while self.get_unpublished_task() is not None:
            pass

    def get_job_to_inject(self, session):
        to_run = or_(
            Job.next_run < datetime.datetime.utcnow(),
            Job.run_now == True,  # noqa
        )
        job = session.query(Job).filter(to_run).first()
        if job is None:
            return None

        logger.info('[cronq_job_id:{0}] Found a job: {1} {2}'.format(
            job.id, job.name, job.next_run))

        return job

    def update_job_time(self, session, job):
        if job.next_run is None:
            current_time = datetime.datetime.utcnow()
            logger.info('[cronq_job_id:{0}] Setting time to {1}'.format(job.id, current_time))
            job.next_run = current_time
        else:
            while job.next_run < datetime.datetime.utcnow():
                logger.info('[cronq_job_id:{0}] Adding time!'.format(job.id))
                job.next_run += job.interval
        job.run_now = False
        me = '{0}.{1}'.format(socket.gethostname(), os.getpid())
        job.locked_by = me

        # update
        try:
            session.merge(job)
            session.commit()
        except InternalError, e:
            session.rollback()
            logger.warning('[cronq_job_id:{0}] Error updating time {1} - {2}'.format(
                job.id, job.name, e))
            return None
        except Exception, e:
            session.rollback()
            logger.exception('[cronq_job_id:{0}] {1} {2}'.format(job.id, job.name, e))
            raise
        else:
            logger.info('[cronq_job_id:{0}] Next job run: {1}'.format(job.id, job.next_run))
            return job

    def get_unpublished_task(self):
        session = self._new_session()

        job = self.get_job_to_inject(session)
        if not job:
            logger.info("no job found")
            session.close()
            return

        # update job time
        job = self.update_job_time(session, job)
        if not job:
            logger.info("no job found after update time")
            session.close()
            return

        # inject
        job_doc = {
            'name': job.name,
            'command': unicode(job.command),
            'id': job.id,
        }

        # publish
        logger.info("[cronq_job_id:{0}] Trying to publish job".format(job.id))
        success = self.publisher.publish(job.routing_key, job_doc, uuid4().hex)
        if not success:
            logger.warning('[cronq_job_id:{0}] Error publishing: {1}'.format(
                job.id, job.name))
            session.close()
            return

        logger.info("[cronq_job_id:{0}] Job published".format(job.id))

        session.close()
        return True

    def event_models_to_docs(self, events):
        for event in events:
            doc = {
                'id': event.id,
                'datetime': event.datetime,
                'run_id': event.run_id,
                'host': event.host,
                'return_code': event.return_code,
                'type': event.type,
                'job_id': event.job_id,
                'status': self.get_status(event.type, event.return_code),
            }
            yield doc

    def get_status(self, _type, return_code=None):
        if _type == self.SUCCEEDED:
            return self.SUCCEEDED
        elif return_code is None:
            return _type
        elif _type == self.FINISHED:
            if int(return_code) == 0:
                return self.SUCCEEDED

        return self.FINISHED


    @staticmethod
    def _add_log_urls(chunks, log_url_template):
        for run_id, run_data in chunks.items():
            log_url = log_url_template \
                .replace('{job_id}', str(run_data['job_id'])) \
                .replace('{run_id}', str(run_id))

            if any([time_field in log_url_template for time_field in ('{start_time}', '{end_time}')]):
                if run_data.get('first'):
                    start_time = urllib.quote(run_data['first']['datetime'].strftime('%Y-%m-%dT%H:%M:%S.000Z'))
                else:
                    # default to now minus 24hrs if field is missing. this generally means bad data
                    logger.warning("No start time found for {0}, using 24 hours ago for log url".format(run_id))
                    one_day_ago = datetime.datetime.utcnow() - datetime.timedelta(hours=24)
                    start_time = urllib.quote(one_day_ago.strftime('%Y-%m-%dT%H:%M:%S.000Z'))

                if run_data.get('last'):
                    end_time = urllib.quote(run_data['last']['datetime'].strftime('%Y-%m-%dT%H:%M:%S.999Z'))
                else:
                    # default to just slightly in the future if not present. probably means still running
                    few_seconds_in_future = datetime.datetime.utcnow() + datetime.timedelta(seconds=5)
                    end_time = urllib.quote(few_seconds_in_future.strftime('%Y-%m-%dT%H:%M:%S.000Z'))
                log_url = log_url.replace('{start_time}', start_time)
                log_url = log_url.replace('{end_time}', end_time)

            run_data['log_url'] = log_url
